//Ported to QNX 4
//1997Feb02(Sun)
//uses fixed size in value table to accomodate messaging
//added optional dump of value table to file
//changes marked with manifest __QNX__
//Dean Z. Douthat
//Osiris Business Systems
//PO Box 7571
//Ann Arbor MI 48107-7571
//1(313)747-9170
//dean@cyberzone-inc.com
//==============================================================================
#include <assert.h>
#include <string.h>
#include <limits.h>
#ifndef __QNX__
#include <values.h>
#else
#include <stdio.h>	//for perror
#include <fcntl.h>	//for open, O_*
#include <unistd.h>	//for basename, write, close
#endif
#include <iostream.h>
#include <fstream.h>
#include <stdlib.h>
#include <math.h>


// Perfect hash finder
// (C) 1996 Carlo Pescio
// Eptacom Consulting
// Via B. Forte 2-3
// 17100 Savona
// ITALY
// FAX +39-19-854761
// email: pescio@mbox.vol.it


// HOW TO USE THIS PROGRAM
// make a data file with the following format:
//   total number of words
//   ... all the words on separated lines
// and pass the filename as a parameter on the command line




// remove this if you want assertions checking
#define NDEBUG


// if not defined from the compiler
typedef int bool ;
const int FALSE = 0 ;
const int TRUE = 1 ;



inline int max( int x, int y )
  {
  return( x > y ? x : y ) ;
  }


class Word
  {
  public :
    Word( const char* s ) ;
    ~Word() ;
    char First() const ;
    char Last() const ;
    unsigned Len() const ;
    bool Conflict( const Word& w ) const ;
  friend ostream& operator <<( ostream& os, const Word& w ) ;
  private :
    char* cstring ;
    unsigned len ;
    char first ;
    char last ;
  } ;


Word :: Word( const char* s )
  {
  len = strlen( s ) ;
  cstring = new char[ len + 1 ] ;
  strcpy( cstring, s ) ;
  first = *s ;
  last = *( s + len - 1 ) ;
  }


Word :: ~Word()
  {
  delete[] cstring ;
  }


inline char Word :: First() const
  {
  return( first ) ;
  }


inline char Word :: Last() const
  {
  return( last ) ;
  }


inline unsigned Word :: Len() const
  {
  return( len ) ;
  }



inline bool Word :: Conflict( const Word& w ) const
  {
  return( first == w.first && last == w.last && len == w.len ) ;
  }


ostream& operator <<( ostream& os, const Word& w )
  {
  os << w.cstring ;
  return( os ) ;
  }


class WordSet
  {
  public :
    WordSet( unsigned maxWords ) ;
    ~WordSet() ;
    void Add( const char* s ) ;
    unsigned Card() const ;
    const Word& operator[]( unsigned index ) const ;

    void CalcFreqFirstLast() ;
    void SortFreqFirstLast() ;
    void Reorder() ;
    int GetMaxLen() const ;
    int GetMinLen() const ;
  private :
    static int CompareFreqs( const void* v1, const void* v2 ) ;
    static WordSet* wsUnderOrdering ;
    Word** set ;
    unsigned card ;
    unsigned maxCard ;
    unsigned freq[ CHAR_MAX ] ;

    void UpdateFreqFirstLast( const Word& w ) ;
  } ;


WordSet* WordSet :: wsUnderOrdering = NULL ;


WordSet :: WordSet( unsigned maxWords )
  {
  set = new Word*[ maxWords ] ;
  for( unsigned i = 0; i < maxWords; i++ )
    set[ i ] = NULL ;
  card = 0 ;
  maxCard = maxWords ;
  for( i = 0; i < CHAR_MAX; i++ )
    freq[ i ] = 0 ;
  }


WordSet :: ~WordSet()
  {
  for( unsigned i = 0; i < card; i++ )
    delete set[ i ] ;
  delete[] set ;
  }


void WordSet :: Add( const char* s )
  {
  assert( card < maxCard ) ;
  assert( set[ card ] == NULL ) ;

  set[ card++ ] = new Word( s ) ;
  }


inline unsigned WordSet :: Card() const
  {
  return( card ) ;
  }


inline const Word& WordSet :: operator[]( unsigned index ) const
  {
  assert( index < card ) ;
  return( *(set[ index ]) ) ;
  }


void WordSet :: CalcFreqFirstLast()
  {
  for( unsigned i = 0; i < card; i++ )
    {
    const Word& w = *(set[ i ]) ;
    UpdateFreqFirstLast( w ) ;
    }
  }


void WordSet :: UpdateFreqFirstLast( const Word& w )
  {
  freq[ w.First() ]++ ;
  freq[ w.Last() ]++ ;
  }


void WordSet :: SortFreqFirstLast()
  {
  // sort words by decreasing sum of frequencies of the first and last char
  wsUnderOrdering = this ;
  qsort( set, card, sizeof( Word* ), WordSet :: CompareFreqs ) ;
  }


int WordSet :: CompareFreqs( const void* v1, const void* v2 )
  {
  const Word* w1 = *((Word**)v1) ;
  const Word* w2 = *((Word**)v2) ;
  unsigned f1 = wsUnderOrdering->freq[ w1->First() ] + wsUnderOrdering->freq[ w1->Last() ] ;
  unsigned f2 = wsUnderOrdering->freq[ w2->First() ] + wsUnderOrdering->freq[ w2->Last() ] ;
  // descending order
  if( f2 < f1 )
    return( -1 ) ;
  else if( f2 > f1 )
    return( 1 ) ;
  else
    return( 0 ) ;
  }


void WordSet :: Reorder()
  {
  // reorder items in the set, so that if an item first and last letter
  // occurred before, the item is moved upward in the list. This heuristic
  // maximize the likelihood that collisions are found early in the
  // assignment of values.

#ifdef __QNX__
  const unsigned NO_OCCURRENCE = SHRT_MAX ;
#else
  const unsigned NO_OCCURRENCE = MAXINT ;
#endif
  for( unsigned i = 1; i < card; i++ )
    {
    const Word& w = *(set[ i ]) ;
    char first = w.First() ;
    char last = w.Last() ;
    unsigned firstOccurredIndex = NO_OCCURRENCE ;
    unsigned lastOccurredIndex = NO_OCCURRENCE ;
    for( unsigned j = 0; j < i; j++ )
      {
      const Word& prev = *(set[ j ]) ;
      if( lastOccurredIndex == NO_OCCURRENCE && ( prev.Last() == last || prev.First() == last ) )
	lastOccurredIndex = j ;
      if( firstOccurredIndex == NO_OCCURRENCE && ( prev.First() == first || prev.Last() == first ) )
	firstOccurredIndex = j ;
      if( firstOccurredIndex != NO_OCCURRENCE &&
	  lastOccurredIndex != NO_OCCURRENCE )
	{
	// both letters occurred before: move the item upward, by
	// pushing down the ones between the maximum occurence
	// index and the item index itself. This preserve the
	// original ordering as much as possible
	unsigned firstToMoveDown = max( firstOccurredIndex, lastOccurredIndex ) + 1 ;
	unsigned lastToMoveDown = i - 1 ;
	Word* itemToMoveUp = set[ i ] ;
	for( unsigned k = lastToMoveDown; k >= firstToMoveDown; k-- )
	  set[ k + 1 ] = set[ k ] ;
	set[ firstToMoveDown ] = itemToMoveUp ;
	break ; // item 'i' handled
	}
      }
    }
  }



int WordSet :: GetMaxLen() const
  {
  int ml = 0 ;
  for( unsigned i = 0; i < card; i++ )
    {
    int l = set[ i ]->Len() ;
    if( l > ml )
      ml = l ;
    }
  return( ml ) ;
  }


int WordSet :: GetMinLen() const
  {
#ifdef __QNX__
  int ml = SHRT_MAX ;
#else
  int ml = MAXINT ;
#endif
  for( unsigned i = 0; i < card; i++ )
    {
    int l = set[ i ]->Len() ;
    if( l < ml )
      ml = l ;
    }
  return( ml ) ;
  }


class SlotTable
  {
  public :
    SlotTable( unsigned size ) ;
    ~SlotTable() ;
    bool IsFree( unsigned slot ) const ;
    void MarkUsed( unsigned slot ) ;
    void MarkFree( unsigned slot ) ;
    unsigned FirstFree() const ;
    unsigned LastFree() const ;
  private :
    enum SlotState { free, used } ;
    SlotState* slots ;
    unsigned firstFree ;
    unsigned lastFree ;
    unsigned size ;
  } ;


SlotTable :: SlotTable( unsigned size )
  {
  slots = new SlotState[ size ] ;
  for( int i = 0; i < size; i++ )
    slots[ i ] = free ;
  firstFree = 0 ;
  lastFree = size - 1 ;
  SlotTable :: size = size ;
  }


SlotTable :: ~SlotTable()
  {
  delete[] slots ;
  }


bool SlotTable :: IsFree( unsigned slot ) const
  {
  return( slots[ slot ] == free ) ;
  }


void SlotTable :: MarkUsed( unsigned slot )
  {
  assert( slot < size ) ;
  slots[ slot ] = used ;

  // this code also guarantee that if no free buckets remain,
  // firstFree > lastFree; this will cause loops to exit immediately
  if( slot == firstFree ) 
    {
    firstFree = size ;
    // must find a next one free (if any)
    for( int i = slot; i < size; i++ )
      if( slots[ i ] == free )
        {
        firstFree = i ;
        break ;
        }
    }
  if( slot == lastFree )
    {
    lastFree = 0 ;
    // must find a previous one free (if any)
    for( int i = slot; i >= 0; i-- )
      if( slots[ i ] == free )
        {
        lastFree = i ;
        break ;
        }
    }
  }


void SlotTable :: MarkFree( unsigned slot )
  {
  slots[ slot ] = free ;
  if( slot < firstFree )
    firstFree = slot ;
  else if( slot > lastFree )
    lastFree = slot ;
  }


unsigned SlotTable :: FirstFree() const
  {
  return( firstFree ) ;
  }


unsigned SlotTable :: LastFree() const
  {
  return( lastFree ) ;
  }



class HashFinder
  {
  public :
    HashFinder( unsigned maxWords, float loadingFactor, 
                bool stopAtFirst = FALSE, bool stopAtMinimal = FALSE ) ;
    ~HashFinder() ;
    void Add( const char* s ) ;
    void FindValues() ;
#ifdef __QNX__
	void setfd( int Fd ) {fd = Fd;}
#endif
  private :
    static const int NO_VALUE ;

    WordSet words ;
    SlotTable buckets ;
#ifdef __QNX__
	int fd;
    short value[ CHAR_MAX ] ;
    short maxTableSize ;
    short minVal ;
    short maxVal ;
#else
    int value[ CHAR_MAX ] ;
    int maxTableSize ;
    int minVal ;
    int maxVal ;
#endif
    float loadingFactor ;
    bool stopFirst ;
    bool stopMinimal ;
    bool stopNow ;

    bool CichelliConditionSatisfied() const ;
    bool Search( unsigned index ) ;
    int Hash( const Word& w ) const ;
  } ;


#ifdef __QNX__
const int HashFinder :: NO_VALUE = SHRT_MAX ;
#else
const int HashFinder :: NO_VALUE = MAXINT ;
#endif


HashFinder :: HashFinder( unsigned maxWords, float loadingFactor,
                          bool stopAtFirst, bool stopAtMinimal ) :
	      words( maxWords ),
              buckets( (int)( maxWords / loadingFactor ) )
  {
  for( unsigned i = 0; i < CHAR_MAX; i++ )
    value[ i ] = NO_VALUE ;

  HashFinder :: loadingFactor = loadingFactor ;
  stopFirst = stopAtFirst ;
  stopMinimal = stopAtMinimal ;
#ifdef __QNX__
	fd = -1;
#endif
  }


HashFinder :: ~HashFinder()
  {
  }

inline void HashFinder :: Add( const char* s )
  {
  words.Add( s ) ;
  }


void HashFinder :: FindValues()
  {
  if( ! CichelliConditionSatisfied() )
    return ;

  words.CalcFreqFirstLast() ;
  words.SortFreqFirstLast() ;
  words.Reorder() ;

#ifdef __QNX__
  maxTableSize = (short)( words.Card() / loadingFactor ) ;
#else
  maxTableSize = (int)( words.Card() / loadingFactor ) ;
#endif

  // use some heuristics to find the range of reasonable values;
  // since we add the lenght, the minimum value must be <= -(max lenght)
  // to allow for positive integers.
  // the maximum value should be <= (maxTableSize-minLen) to fit into the
  // table. Note that max value is excluded in the loops.
  // To keep the range small, the range is set as follows, and is then
  // refined as one of the two gets assigned (see Search() below)
#ifdef __QNX__
  minVal =(short) -( words.GetMaxLen() ) ;
  maxVal = (short)(maxTableSize - words.GetMinLen()) ;
#else
  minVal = -( words.GetMaxLen() ) ;
  maxVal = maxTableSize - words.GetMinLen() ;
#endif

  stopNow = FALSE ;
  if( ! Search( 0 ) )
    cerr << "Cannot find perfect hashing\n\n" ;
  }


bool HashFinder :: CichelliConditionSatisfied() const
  {
  unsigned card = words.Card() ;
  for( unsigned i = 0; i < card; i++ )
    for( unsigned j = i + 1; j < card; j++ )
      if( words[ i ].Conflict( words[ j ] ) )
	{
	cerr << "Word set does not satisfy Cichelli's condition\n" ;
	cerr << "Conflicting pair: " << words[ i ] << " , " << words[ j ] ;
	cerr << "\nProcess aborted\n\n" ;
	return( FALSE ) ;
        }
  return( TRUE ) ;
  }


inline int HashFinder :: Hash( const Word& w ) const
  {
  return( value[ w.First() ] + value[ w.Last() ] + w.Len() ) ;
  }

bool HashFinder :: Search( unsigned index )
  {
  const Word& w = words[ index ] ;
  if( value[ w.First() ] != NO_VALUE && value[ w.Last() ] != NO_VALUE )
    {
    // try to insert into the table
    int h = Hash( w ) ;
    if( h < 0 || h >= maxTableSize || ! buckets.IsFree( h ) )
      return( FALSE ) ;
    else
      {
      buckets.MarkUsed( h ) ;
      index++ ;
      int card = words.Card() ;
      if( index == card )
	{
#ifdef __QNX__
	if (fd != -1)
		if (write(fd, value, CHAR_MAX*sizeof(short)) != CHAR_MAX*sizeof(short))
		{
			perror("writing output file");
			exit(EXIT_FAILURE);
		}
#endif
	cout << "------------\n" ;
	cout << "perfect hashing found\n" ;
	for( int i = 0; i < CHAR_MAX; i++ )
	  if( value[ i ] != NO_VALUE )
	    cout << (char)i << " = " << value[ i ] << "  " ;
	cout << endl ;
	for( i = 0; i < card; i++ )
	  cout << words[ i ] << " = " << Hash( words[ i ] ) << endl ;
        // calculate and print loading factor
        int used = 0 ;
        for( i = 0; i < maxTableSize; i++ ) 
          if( ! buckets.IsFree( i ) )
            used++ ;
        cout << "loading factor: " << (float)used/maxTableSize << endl ;
        
        // decide if it is time to stop now
        stopNow = stopFirst || ( used == maxTableSize && stopMinimal ) ;

	buckets.MarkFree( h ) ; // restore for backtracking
	return( TRUE ) ;
	}
      else
	{
	bool res = Search( index ) ;
	buckets.MarkFree( h ) ; // restore for backtracking
	return( res ) ;
	}
      }
    }
  else if( value[ w.First() ] == NO_VALUE && value[ w.Last() ] == NO_VALUE )
    {
    // must find values for both first and last; 

    int firstValToAssign = w.First() ;
    int lastValToAssign = w.Last() ;
    bool found = FALSE ;
#ifdef __QNX__
    for( short i = minVal; i < maxVal && ! stopNow; i++ )
#else
    for( int i = minVal; i < maxVal && ! stopNow; i++ )
#endif
      {
      value[ firstValToAssign ] = i ;
      // here we can change the heuristics: we already assigned one of
      // the values, so we have modified the boundaries.
      // Now we must have FirstFree <= to-assign + 'i' + len <= LastFree
#ifdef __QNX__
      short minVal2 = (short)(buckets.FirstFree() - i - w.Len()) ;
      short maxVal2 = (short)(buckets.LastFree() - i - w.Len() + 1) ; // max not reached
      for( short j = minVal2; j < maxVal2 && ! stopNow; j++ )
#else
      int minVal2 = buckets.FirstFree() - i - w.Len() ;
      int maxVal2 = buckets.LastFree() - i - w.Len() + 1 ; // max not reached
      for( int j = minVal2; j < maxVal2 && ! stopNow; j++ )
#endif
	{
	value[ lastValToAssign ] = j ;
	bool res = Search( index ) ; // will take first branch this time
	found = found ||  res ;
	}
      }
    value[ firstValToAssign ] = NO_VALUE ; // restore for backtracking
    value[ lastValToAssign ] = NO_VALUE ;
    return( found ) ;
    }
  else
    {
    // must find valid values for the one without values
    int valToAssign ;
    int valAssigned ;
    if( value[ w.First() ] == NO_VALUE )
      {
      valToAssign = w.First() ;
      valAssigned = w.Last() ;
      }
    else
      {
      valToAssign = w.Last() ;
      valAssigned = w.First() ;
      }

    // see above for heuristics used here
#ifdef __QNX__
    short minVal2 = (short)(buckets.FirstFree() - value[valAssigned] - w.Len());
    short maxVal2 = (short)(buckets.LastFree()-value[valAssigned]-w.Len() + 1);
#else
    int minVal2 = buckets.FirstFree() - value[ valAssigned ] - w.Len() ;
    int maxVal2 = buckets.LastFree() - value[ valAssigned ] - w.Len() + 1 ;
#endif
    assert( value[ valToAssign ] == NO_VALUE ) ;
    bool found = FALSE ;
#ifdef __QNX__
    for( short i = minVal2; i < maxVal2 && ! stopNow; i++ )
#else
    for( int i = minVal2; i < maxVal2 && ! stopNow; i++ )
#endif
      {
      value[ valToAssign ] = i ;
      bool res = Search( index ) ;  // will take first branch this time
      found = found ||  res ; 
      }
    value[ valToAssign ] = NO_VALUE ; // restore for backtracking
    return( found ) ;
    }
  }


bool ParseChoice( char ch )
  {
  switch( ch )
    {
    case 'y' :
    case 'Y' :
      return( TRUE ) ;
    case 'n' :
    case 'N' :
      return( FALSE ) ;
    default :
      assert( FALSE ) ;
      return( FALSE ) ;
    }
  }

int main( int argc, char* argv[] )
  {
#ifdef __QNX__
  if( argc < 2 )
    {
    cerr << "syntax: "<<basename(argv[0])<<" infile [outfile]"<< endl;
#else
  if( argc != 2 )
    {
    cerr << "syntax: perfhash infile\n" ;
#endif
    return( 1 ) ;
    }
  
  cout << "loading factor (0,1] : " ;
  float lf = -1 ;
  cin >> lf ;
  assert( 0 < lf && lf <= 1 ) ;

  cout << "stop at first solution (Y/N) ? " ;
  char ch ;
  cin >> ch ;
  bool stopFirst = ParseChoice( ch ) ;

  bool stopFirstMinimal = FALSE ;
  if( ! stopFirst )
    {
    cout << "stop at first minimal solution (Y/N) ? " ;
    cin >> ch ;
    stopFirstMinimal = ParseChoice( ch ) ;
    }

  ifstream is( argv[ 1 ] ) ;
  int size = 0 ;
  is >> size ;
  assert( size > 0 ) ;

  HashFinder hf( size, lf, stopFirst, stopFirstMinimal ) ;
  
#ifdef __QNX__
	int Fd = -1;
	if (argc >= 3 && (Fd = open(argv[2], O_WRONLY|O_CREAT|O_TRUNC, 0664)) == -1)
	{
		perror("Opening output file");
		exit(EXIT_FAILURE);
	}
	hf.setfd(Fd);
#endif
  const unsigned MAX_WORD_LEN = 256 ;
  char buf[ MAX_WORD_LEN ] ;
  for( int i = 0; i < size; i++ )
    {
    is >> buf ;
    hf.Add( buf ) ;
    }

  hf.FindValues() ;


  return( 0 ) ;
  }
